--[[
	author: Zax
	content:
		globales: ZXSNDLVL_
		functions: SndMgr.
		XML: 
		XML UI: 
		main UI frame: 
		event frame: SoundManagerFrame
		XML config: 
	notes: Classic compatible
--]]

---------------------------------------------------------------------------------------
------------------------------------- VARIABLES ---------------------------------------
SndMgr = {}; -- nommage espace
local addonName = ...

local addonVar = {
	["name"] = addonName, -- nécessaire
	["version"] = tonumber(GetAddOnMetadata(addonName, "Version")), -- nécessaire
	["chatColor"] = "80FFFF", -- nécessaire
	["slashCmd"] = "soundmanager", -- nécessaire
	["settingsPanel"] = "", -- nom du frame de settings : laisser vide si utilisation du frame de WoW (ou si pas de settings)
	
	["btnWidth"] = 30, -- px
	["btnHeight"] = 23, -- px
	["smlBtnWidth"] = 15, -- px
	["alphaUI"] = 0.5,
	["textdelay"] = 3, -- delai d'affichage du device (secondes)
}

function SndMgr.getLocal(key)
	----------- pour traduction (voir localization.lua) : ne pas deplacer !!!
	return key
end

-------------------------------------------- saved et init
ZXSNDMGR_SAVED = {};

function SndMgr.initData()
	local tbl = {
		["settings"] = {
			["version"] = GetAddOnMetadata(addonVar["name"],"Version"),
			["debug"] = 0,
			["visible"] = 1,
			["framebtn"] = {
				["posX"] = 200,
				["posY"] = 600, -- distance bas UIParent / haut main frame
			},
			["initialDevice"] = "", -- pas NIL ! : device chargé au premier login (et pas au rechargement d'UI)
			["lastLogout"] = 0, -- pas NIL ! : date du dernier "vrai" login
		},
		["outputs"] = {
			["outputNameToIgnore"] = {},
			["volumes"] = {}, -- une table par device avec le nom en clé et le volume en valeur
		},
	}
	return tbl
end

local soundEventTbl = { -- liste des events spéciaux gérés pour mise à jour du frame
	"CHAT_MSG_CHANNEL",
	"CHAT_MSG_SAY",
	"CHAT_MSG_PARTY",
	"CHAT_MSG_RAID",
	"CHAT_MSG_GUILD",
	"CHAT_MSG_EMOTE",
	"CHAT_MSG_MONSTER_SAY",
	"CHAT_MSG_MONSTER_EMOTE",
}

local soundDevicesTbl = {} -- liste de tous les output devices
local logoutDefaultDelay = 60*60*2 -- 2 heures de délais pour considérer un vrai login (et non pas un rechargement de l'UI)

---------------------------------------------------------------------------------------
--------------------------------------- EVENTS ----------------------------------------
function SndMgr.eventHandler(self, event, ...)
	if (event == "PLAYER_LOGIN") then
		SndMgr.loading()
	elseif (event == "PLAYER_ENTERING_WORLD") then
		SndMgr.enteringWorld(...) -- spécial pour différencier un premier lancement d'un simple rechargement de l'UI
	elseif (event == "PLAYER_LOGOUT") then
		ZXSNDMGR_SAVED["settings"]["lastLogout"] = time() -- secondes
	elseif (tContains(soundEventTbl, event)) then
		-- tous les autres events "spéciaux" ne servent qu'à mettre à jour l'affichage du sound volume
SndMgr.testPrint("Event="..event)--/////
		SndMgr.updateSoundFrame(false)
	end          
end

function SndMgr.enteringWorld(isInitialLogin, isReloadingUi)
	if (isInitialLogin or isReloadingUi) then
		if (isInitialLogin) then
--print("PLAYER_ENTERING_WORLD: first login !!!!!!!!!!!") --/////
			if ((time() - ZXSNDMGR_SAVED["settings"]["lastLogout"]) > logoutDefaultDelay ) then
--print("PLAYER_ENTERING_WORLD: délai écoulé --> switchToDefaultDevice") --/////
				SndMgr.switchToDefaultDevice()
				SndMgr.updateSoundFrame(true)
				SndMgr.showDeviceNameFrame(nil, true)
			end
		end
	else
		-- changement de zones instanciées : ne rien faire
	end
end

---------------------------------------------------------------------------------------
--------------------------------- CHARGEMENT & BINDINGS -------------------------------
function SndMgr.loading()
	------------------------------- slash commands
	SlashCmdList["ZXSNDMGR_CMD"] = SndMgr.slashCmd
		SLASH_ZXSNDMGR_CMD1 = "/soundmanager"
		SLASH_ZXSNDMGR_CMD2 = "/sndmanager"
		SLASH_ZXSNDMGR_CMD3 = "/soundmgr"
		SLASH_ZXSNDMGR_CMD4 = "/sndmgr"
				
	------------------------------- bindings
	BINDING_HEADER_ZXSNDMGR_TITLE = addonName
	BINDING_NAME_ZXSNDMGRPLUS = SndMgr.getLocal("Increase sound volume")
	BINDING_NAME_ZXSNDMGRMINUS = SndMgr.getLocal("Decrease sound volume")
	BINDING_NAME_ZXSNDMGRSWITCH = SndMgr.getLocal("Switch to next device")

	------------------------------- settings general
	if (SndMgr.isEmpty(ZXSNDMGR_SAVED)) then
		ZXSNDMGR_SAVED = SndMgr.initData()
--print("=== no general lua file : INIT SETTINGS ===") --/////
	else
		local tbl = SndMgr.initData()
		SndMgr.setSavedCompat120()
		if (SndMgr.tableKeyDiff(tbl, ZXSNDMGR_SAVED, "") ~= nil) then
			ZXSNDMGR_SAVED = SndMgr.copyTable(tbl, false)
--print("=== diff general saved data : INIT SETTINGS ===")--/////
		end
	end
	
	------------------------------- suite
	SndMgr.ui_createSoundFrame()
	soundDevicesTbl = SndMgr.getOuputDevicesNames() -- construction de la liste des devices (ne pas déplacer !!!)
	if (SndMgr.isTrue(ZXSNDMGR_SAVED["settings"]["visible"])) then
		SndMgr.showHide(true)
		SndMgr.listenEvents(true)
		SndMgr.updateSoundFrame(true)
		SndMgr.showDeviceNameFrame(nil, true)
	end
	
	local str = ""
	if (not SoundManagerFrame:IsVisible()) then str = " "..SndMgr.getLocal("Type /soundmanager to display the addon interface.") end
	if (SndMgr.debug()) then
		SetCVar("scriptErrors", 1)
		str = str..SndMgr.getTxtColor(" ------ DEBUG ON.", "orange")
	end
	if (str ~= "") then SndMgr.echoChatAppl(str, true, nil) end
end

function SndMgr.setSavedCompat120() -- pour ne pas reset les prefs de la 1.20 avec les versions suivantes
	if (ZXSNDMGR_SAVED["settings"]["initialDevice"] == nil) then ZXSNDMGR_SAVED["settings"]["initialDevice"] = "" end
	if (ZXSNDMGR_SAVED["settings"]["lastLogout"] == nil) then ZXSNDMGR_SAVED["settings"]["lastLogout"] = date("%Y.%m.%d", GetServerTime()) end
end

---------------------------------------------------------------------------------------
-------------------------------------- SLASH CMD --------------------------------------
function SndMgr.slashCmd(arg, execLevel)
	SndMgr.slashCmd_Lib(arg, execLevel)
end

function SndMgr.slashCmd_getList(arg2) -- liste des commandes-slash (voir .lib)
	-- .slashCmd_showList(dontPrint) -- print la liste en print, ou la renvoit en chaine avec dontPrint à true (pour frame cfg)
	local tbl = {
		["user"] = {
			{
				["cmd"] = {"show"},
				["func"] = function() SndMgr.showHide(true) end,
				["label"] = SndMgr.getLocal("show addon interface"),
			},
			{
				["cmd"] = {"hide"},
				["func"] = function() SndMgr.showHide(false) end,
				["label"] = SndMgr.getLocal("hide addon interface"),
			},
			{
				["cmd"] = {"up", "+"},
				["func"] = function() SndMgr.increaseVolume() end,
				["label"] = SndMgr.getLocal("increase sound volume"),
			},
			{
				["cmd"] = {"down", "-"},
				["func"] = function() SndMgr.decreaseVolume() end,
				["label"] = SndMgr.getLocal("decrease sound volume"),
			},
			{
				["cmd"] = {"switch"},
				["func"] = function() SndMgr.execOutputRotation() end,
				["label"] = SndMgr.getLocal("switch to next non-ignored sound output device"),
			},
			{
				["cmd"] = {"ignore"},
				["func"] = function() SndMgr.ignoreDevice() end,
				["label"] = SndMgr.getLocal("current device will be ignored when changing sound output device"),
			},
			{
				["cmd"] = {"noignore"},
				["func"] = function() SndMgr.ignoreNoDevice() end,
				["label"] = SndMgr.getLocal("no output device will be ignored"),
			},
			{
				["cmd"] = {"settings"},
				["func"] = function() SndMgr.showSettings() end,
				["label"] = SndMgr.getLocal("display current addon settings"),
			},
--[[ §§§
			{
				["cmd"] = {"place", "map", "minimap"},
				["func"] = function() SndMgr.placeToMap() end,
				["label"] = SndMgr.getLocal("move SoundManager frame near minimap name"),
			},
--]]
			
			{
				["cmd"] = {"reset", "resetall"},
				["func"] = function() SndMgr.resetall() end,
				["label"] = SndMgr.getLocal("reset all addon settings to default"),
			},
			{
				["cmd"] = {"help", "aide", "?"},
				["func"] = function() SndMgr.help() end,
				["label"] = SndMgr.getLocal("display available slash commands"),
			},
		},
		["dev"] = {
			{
				["cmd"] = {"test"},
				["func"] = function() SndMgr.fcnTest(arg2) end,
				["label"] = "DEV test with 1 argument",
			},
			{
				["cmd"] = {"debug"},
				["func"] = function() SndMgr.toggleDebug() end,
				["label"] = "toggle debug",
			},
			{
				["cmd"] = {"showhide"}, -- pour DEFAULTCMD
				["func"] = function() SndMgr.showHide() end,
				["label"] = SndMgr.getLocal("show addon interface"),
			},
		},
		["DEFAULTCMD"] = "showhide", -- commande par défaut
		["DEFAULTARG"] = "help", -- argument par défaut d'une commande
	}
	return tbl
end

---------------------------------------------------------------------------------------
--------------------------------------- ACTIONS ---------------------------------------
function SndMgr.help()
	SndMgr.slashCmd_showList(false) -- pas d'UI (mettre à true pour ne pas printer la liste)
end

function SndMgr.fcnTest(arg) -- $$$
	--print(arg)
	print(date("%Y.%m.%d %H:%M:%S", GetServerTime()))
	local version, build, datevers, tocversion = GetBuildInfo()
	print(version, build, datevers, "toc = "..tocversion)
	print("------------------------------------------")
	--print("Sound_MasterVolume", GetCVar("Sound_MasterVolume"))
end

function SndMgr.resetall()
	wipe(ZXSNDMGR_SAVED)
	ZXSNDMGR_SAVED = SndMgr.initData()
	soundDevicesTbl = SndMgr.getOuputDevicesNames() -- reconstruction de la liste des devices
	SndMgr.echoChatAppl("settings reset to default.", false, "yellow")
	SndMgr.updateSoundFrame(true)
	SoundManagerFrame:ClearAllPoints() -- Clears any set anchors.
	SoundManagerFrame:SetPoint("CENTER", UIParent)
	SndMgr.saveFramesPosition()
	SndMgr.showHide(true)
end

function SndMgr.showHide(optionalBoolean)
	local val, userArg
	if (optionalBoolean == nil) then
		----------- switch
		val = (not SoundManagerFrame:IsVisible())
	else
		----------- given value
 		val = SndMgr.isTrue(optionalBoolean, true)
		userArg = val
	end
	
	ZXSNDMGR_SAVED["settings"]["visible"] = SndMgr.booleanToInt(val)
	SndMgr.listenEvents(val)
	if (val) then
		SndMgr.updateSoundFrame(true)
		SoundManagerFrame:Show()
		SndMgr.showDeviceNameFrame(nil, true)
	else
		SoundManagerFrame:Hide()
		if (not SndMgr.isTrue(userArg)) then SndMgr.echoChatAppl(SndMgr.getLocal("hidden. Type /soundmanager to show interface again."), false, "yellow") end -- rappel sur action user
	end
end

function SndMgr.toggleDebug() -- pour dev
	if (SndMgr.isTrue(ZXSNDMGR_SAVED["settings"]["debug"])) then ZXSNDMGR_SAVED["settings"]["debug"] = 0 else ZXSNDMGR_SAVED["settings"]["debug"] = 1 end
	SndMgr.echoChatAppl("debug "..tostring(SndMgr.debug())..".", false, nil)
end

function SndMgr.showSettings()
	SndMgr.echoChatAppl("account settings:", false, nil)
	SndMgr.echoTable(ZXSNDMGR_SAVED, "   ")
end

function SndMgr.ignoreDevice(deviceName) -- nom complet du device à ignorer (ou actuel si vide)
	local deviceName = deviceName or SndMgr.getCurrentDeviceName()
	--------------------------- ajouter aux ignored
	if (not SndMgr.isTrue(tContains(ZXSNDMGR_SAVED["outputs"]["outputNameToIgnore"], deviceName))) then
		tinsert(ZXSNDMGR_SAVED["outputs"]["outputNameToIgnore"], deviceName)
		soundDevicesTbl = SndMgr.getOuputDevicesNames() -- reconstruction de la liste des devices
	end
--SndMgr.echoDebug(ZXSNDMGR_SAVED["outputs"]["outputNameToIgnore"], ".ignoreDevice(): SAVED") -- /////
--SndMgr.echoDebug(soundDevicesTbl, ".ignoreDevice(): soundDevicesTbl") -- /////
	SndMgr.echoChatAppl("'"..deviceName.."' "..SndMgr.getLocal("will be ignored."), false, "orange")
end

function SndMgr.activeDevice(deviceName) -- nom complet du device à activer (ou actuel si vide)
	local deviceName = deviceName or SndMgr.getCurrentDeviceName()
	--------------------------- virer des ignored
	for i, value in ipairs(ZXSNDMGR_SAVED["outputs"]["outputNameToIgnore"]) do
		if (value == deviceName) then
			table.remove(ZXSNDMGR_SAVED["outputs"]["outputNameToIgnore"], i)
			break
		end
	end
	soundDevicesTbl = SndMgr.getOuputDevicesNames() -- reconstruction de la liste des devices
--SndMgr.echoDebug(soundDevicesTbl, ".activeDevice()") -- /////
	SndMgr.echoChatAppl("'"..deviceName.."' "..SndMgr.getLocal("is no longer ignored."), false, "green")
end

function SndMgr.ignoreNoDevice() -- aucun device ne sera ignoré (vide la liste des ignorés)
	wipe(ZXSNDMGR_SAVED["outputs"]["outputNameToIgnore"])
	soundDevicesTbl = SndMgr.getOuputDevicesNames() -- reconstruction de la liste des devices
	SndMgr.echoChatAppl("no output device will be ignored", false, "orange")
end

function SndMgr.increaseVolume()-- SetCVar("Sound_MasterVolume"), 0 to 1 --> 0.5= 50%)
	local currentLvl = GetCVar("Sound_MasterVolume")
	local newLvl = SndMgr.convertToPercent(currentLvl) + 10
	newLvl = min(100, newLvl/100)
	SetCVar("Sound_MasterVolume", newLvl)

	SndMgr.updateSoundFrame(true)
end

function SndMgr.decreaseVolume()-- SetCVar("Sound_MasterVolume"), 0 to 1 --> 0.5= 50%)
	local currentLvl = GetCVar("Sound_MasterVolume")
	local newLvl = SndMgr.convertToPercent(currentLvl) - 10
	newLvl = max(0, newLvl/100)
	SetCVar("Sound_MasterVolume", newLvl)

	SndMgr.updateSoundFrame(true)
end

function SndMgr.placeToMap() -- UNUSED wow10
	SoundManagerFrame:ClearAllPoints()
	SoundManagerFrame:SetPoint("TOP", "Minimap", "BOTTOM", 0, -10)
	SndMgr.saveFramesPosition()
end

---------------------------------------------------------------------------------------
------------------------------------ FCNS INTERNES ------------------------------------

function SndMgr.getOuputDevicesNames() -- renvoit table associative de tous les devices, ignorés et actifs
	-- la construction des ignorés se fait par rapport à la table ZXSNDMGR_SAVED["outputs"]["outputNameToIgnore"]
	local tbl = {
		["active"] = {},
		["ignored"] = {},
	}
	
	local thisName = ""
	local thisIgnored = 0
	for i = 1, (Sound_GameSystem_GetNumOutputDrivers() - 1) do -- renvoit 1 de plus (pour "aucun")
		thisName = Sound_GameSystem_GetOutputDriverNameByIndex(i)
		if (SndMgr.isTrue(tContains(ZXSNDMGR_SAVED["outputs"]["outputNameToIgnore"], thisName))) then
			tinsert(tbl.ignored, {["index"] = i, ["name"] = SndMgr.getSystemName(thisName)})
		else
			tinsert(tbl.active, {["index"] = i, ["name"] = SndMgr.getSystemName(thisName)})
		end
	end
--SndMgr.echoDebug(tbl, "getOuputDevicesNames()") --/////
	return tbl
end

function SndMgr.getSystemName(name) -- UNUSED : tentative de nommer l'option "Systeme" (réglable dans le jeu)
	local deviceName = name or ""
	if (deviceName == "") then
		return SndMgr.getLocal("System")
	else
		if (tContains({"none", "aucun",}, strlower(deviceName))) then return SndMgr.getLocal("System") else return deviceName end
	end
end

function SndMgr.outputRotation() -- passe au device audio suivant et renvoit son nom (str), ou nil
	local currentOuputIndex = tonumber(GetCVar("Sound_OutputDriverIndex")) -- string ?
	local indexTblFound = 0
	for i, value in ipairs(soundDevicesTbl.active) do
		if (value.index == currentOuputIndex) then
			indexTblFound = i
			break
		end
	end
if (SndMgr.debug()) then
	SndMgr.echoDebug(soundDevicesTbl, "outputRotation(): soundDevicesTbl")
	print("outputRotation(): ---currentOuputIndex=", currentOuputIndex, "---indexTblFound=", indexTblFound)
end
	if (indexTblFound > 0) then
		if (indexTblFound + 1 > #soundDevicesTbl.active) then indexTblFound = 1 else indexTblFound = indexTblFound + 1 end
	else
		indexTblFound = 1
	end
	SetCVar("Sound_OutputDriverIndex", soundDevicesTbl.active[indexTblFound].index)
	
	local success, errorMessage = pcall(Sound_GameSystem_RestartSoundSystem)
	if (not success) then AudioOptionsFrame_AudioRestart() end -- non wow10
	
	return soundDevicesTbl.active[indexTblFound].name
end

function SndMgr.getCurrentDeviceName() -- renvoit le nom de l'output device courant
	local deviceIndex = tonumber(GetCVar("Sound_OutputDriverIndex"))
	return Sound_GameSystem_GetOutputDriverNameByIndex(deviceIndex)
end

function SndMgr.switchToDefaultDevice() -- au premier lancement
	if (not SndMgr.isEmpty(ZXSNDMGR_SAVED["settings"]["initialDevice"])) then
		if (SndMgr.getCurrentDeviceName() ~= ZXSNDMGR_SAVED["settings"]["initialDevice"]) then
			local newDeviceName = ZXSNDMGR_SAVED["settings"]["initialDevice"]
			local indexTblFound = 0
			for i, value in ipairs(soundDevicesTbl.active) do
				if (value.name == newDeviceName) then
					indexTblFound = i
					break
				end
			end
			if (indexTblFound > 0) then
				--------------- switch device
				SndMgr.testPrint("switchToDefaultDevice() : indexTblFound="..indexTblFound.."---deviceName="..soundDevicesTbl.active[indexTblFound].name.."---")
				SetCVar("Sound_OutputDriverIndex", soundDevicesTbl.active[indexTblFound].index)
	
				local success, errorMessage = pcall(Sound_GameSystem_RestartSoundSystem)
				if (not success) then AudioOptionsFrame_AudioRestart() end -- non wow10
		
				--------------- ajustement volume
				SndMgr.setVolumeSavedDevice(newDeviceName)
			else
				SndMgr.testPrint("switchToDefaultDevice() : indexTblFound="..indexTblFound.."---> DEVICE NOT FOUND")
				ZXSNDMGR_SAVED["settings"]["initialDevice"] = ""
			end
		end
	end
end

function SndMgr.setVolumeSavedDevice(savedDeviceName) -- applique le niveau sonore sauvegardé pour un nom de device et renvoit le nouveau volume sonore
	local changeVolumeStr = ""
	local volumeDevice = ZXSNDMGR_SAVED["outputs"]["volumes"][savedDeviceName]
	if (volumeDevice ~= nil) then
		volumeDevice = tonumber(volumeDevice)
		if (volumeDevice ~= nil) then
			SetCVar("Sound_MasterVolume", volumeDevice)
			changeVolumeStr = SndMgr.convertToPercent(volumeDevice)
		end
	end
	return changeVolumeStr
end

function SndMgr.setDefaultDevice(deviceName) -- commande menu local
	local deviceName = deviceName or SndMgr.getCurrentDeviceName()
	ZXSNDMGR_SAVED["settings"]["initialDevice"] = deviceName
	SndMgr.echoChatAppl("'"..deviceName.."' "..SndMgr.getLocal("set to default."), false, "green")
end

function SndMgr.setNoDefaultDevice(deviceName) -- commande menu local
	ZXSNDMGR_SAVED["settings"]["initialDevice"] = ""
	SndMgr.echoChatAppl(SndMgr.getLocal("Default device reset."), false, "green")
end

function SndMgr.saveFramesPosition()
	ZXSNDMGR_SAVED["settings"]["framebtn"]["posX"] = floor(SndMgr.strToNumber(SoundManagerFrame:GetLeft()))
	ZXSNDMGR_SAVED["settings"]["framebtn"]["posY"] = floor(SndMgr.strToNumber(SoundManagerFrame:GetTop())) -- distance entre bas UIParent (ou screen ?) / haut main frame
end

function SndMgr.listenEvents(boolean)
	if (SndMgr.isTrue(boolean)) then
		for i, value in ipairs(soundEventTbl) do
			SoundManagerFrame:RegisterEvent(value)
		end
	else
		for i, value in ipairs(soundEventTbl) do
			SoundManagerFrame:UnregisterEvent(value)
		end
	end
end

---------------------------------------------------------------------------------------
------------------------------------- LIB DEDIEE --------------------------------------

function SndMgr.debug() -- raccourci : renvoit true si debug ON
	return SndMgr.isTrue(ZXSNDMGR_SAVED["settings"]["debug"])
end

function SndMgr.getVar(nomVar, subLvl1, subLvl2) -- nécessite addonVar
	-- get valeurs de la table locale addonVar
	-- si aucun argument n'est passé, renvoit la table addonVar
	if (addonVar == nil) then return nil end
	
	if (subLvl2 ~= nil) then
		if (addonVar[nomVar][subLvl1][subLvl2] ~= nil) then return addonVar[nomVar][subLvl1][subLvl2] end
	elseif (subLvl1 ~= nil) then
		if (addonVar[nomVar][subLvl1] ~= nil) then return addonVar[nomVar][subLvl1] end
	elseif (nomVar ~= nil) then
		if (addonVar[nomVar] ~= nil) then return addonVar[nomVar] end
	else
		return addonVar
	end
end

function SndMgr.setVar(nomVar, valeur) -- nécessite addonVar
	if (addonVar == nil) then return nil end
	if (nomVar ~= nil) then
		if (addonVar[nomVar] ~= nil) then addonVar[nomVar] = valeur end
	end
end

function SndMgr.getSavedSettings(perCharacterBoolean) -- pour appel depuis autres fichiers
	if (perCharacterBoolean) then return ZXSNDMGR_SAVED else return ZXSNDMGR_SAVED end
end

function SndMgr.convertToPercent(value) -- conversion en % (value entre 0 et 1)
	local value = min(1, SndMgr.strToNumber(value))
	return (SndMgr.round(value, 1, false) * 100)
end

---------------------------------------------------------------------------------------
------------------------------------------ UI -----------------------------------------

function SndMgr.ui_createSoundFrame() -- modif main frame
	--local SoundManagerFrame = CreateFrame("Frame", UIParent, self, BackdropTemplateMixin and "BackdropTemplate")
	---------------------------------------------------------- main frame : fond
	SoundManagerFrame:SetFrameStrata("HIGH") -- pour passer au dessus des autres frames courants
	SoundManagerFrame:SetClampedToScreen("true")
	--SoundManagerFrame:ClearAllPoints()
	SoundManagerFrame:SetMovable(true)
 	SoundManagerFrame:SetUserPlaced(true)
	SoundManagerFrame:SetSize(addonVar.btnWidth + (addonVar.smlBtnWidth * 2), addonVar.btnHeight)
	--SoundManagerFrame:SetPoint("TOPLEFT", "$parent", "TOPLEFT", 30, -300)
	SoundManagerFrame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", floor(SndMgr.nombreParDefaut(ZXSNDMGR_SAVED["settings"]["framebtn"]["posX"], 100, 1, 5000)), floor(SndMgr.nombreParDefaut(ZXSNDMGR_SAVED["settings"]["framebtn"]["posY"], 600, 1, 5000)))
	--SoundManagerFrame:Show() -- voir loading()
	
	
	if (not SoundManagerFrame.SetBackdrop) then Mixin(SoundManagerFrame, BackdropTemplateMixin) end
	SoundManagerFrame:SetBackdrop({
		--bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		tile = false,
		tileSize = 8,
		edgeSize = 8,
			insets = {left = 1, right = 1, top = 1, bottom = 0},
	})
	local r, g, b, alpha = SndMgr.ui_getBkColor(false)
	SoundManagerFrame:SetBackdropColor(r, g, b, alpha)
	
	--------------------------------------------------------- init POPUP MENU
	local framePopupMenu = SndMgr.ui_initializePopupMenu()
	
	--------------------------------------------------------- bouton principale
	local btnMain = CreateFrame("Button", "SoundManager_Switch", SoundManagerFrame)
	btnMain:SetSize(addonVar.btnWidth, addonVar.btnHeight - 6)
	btnMain:SetPoint("CENTER", SoundManagerFrame, "CENTER", 0, 0)
	--------------------------------- texture
	SndMgr.setColoredButton(btnMain, SndMgr.ui_getBkColor(false), nil, 0, addonVar.alphaUI, 0.2)
	--------------------------------- texte
	btnMain.text = btnMain:CreateFontString("$parent_Text", "OVERLAY", "GameFontNormalSmall")
	SndMgr.setFontStyle(btnMain.text, SndMgr.getStuffPath("Myriad Condensed Web.ttf", true), 11, "yellow", false, 1)
	btnMain.text:SetPoint("CENTER", "$parent", "CENTER", 0, 0)
	btnMain.text:SetJustifyH("CENTER")
	btnMain.text:SetText("0")--/////
	--------------------------------- tooltip
    btnMain:SetScript("OnEnter", function(self)
		local r, g, b, alpha = SndMgr.ui_getBkColor(true)
		btnMain.texture:SetColorTexture(r, g, b, alpha)
		GameTooltip:SetOwner(self, "ANCHOR_LEFT", -17, -18)
		GameTooltip:AddLine("Master Sound Volume".."\n".."     Click to switch to next available output device.".."\n".."     Right-click to display more options.".."\n".."     Drag with shift or alt to move this frame.")
		GameTooltip:Show()
		SndMgr.showDeviceNameFrame(nil, false)
    end)
    btnMain:SetScript("OnLeave", function(self)
		local r, g, b, alpha = SndMgr.ui_getBkColor(false)
		btnMain.texture:SetColorTexture(r, g, b, alpha)
		GameTooltip:Hide()
		SndMgr.hideDeviceNameFrame()
	end)
	--------------------------------- drag déplacement et clicks
	btnMain:SetScript("OnMouseDown", function(self, mouse)
		if ((mouse == "LeftButton" and IsModifierKeyDown()) and not SoundManagerFrame.isMoving) then
			SoundManagerFrame:StartMoving()
			SoundManagerFrame.isMoving = true
		end
	end)
	btnMain:SetScript("OnMouseUp", function(self, mouse)
		if (SoundManagerFrame.isMoving) then
			SoundManagerFrame:StopMovingOrSizing()
			SoundManagerFrame.isMoving = false
			SndMgr.saveFramesPosition()
		else
			if (mouse == "LeftButton") then
				----------------- gauche : rotation
				SndMgr.flash(SoundManager_Switch.texture, 1)
				SndMgr.execOutputRotation()
			else
				----------------- droite : appel POPUP MENU
				ToggleDropDownMenu(1, nil, framePopupMenu, "SoundManager_Switch", -20, 1)
			end
		end
	end)

	--------------------------------------------------------- mini btns gauche "-" et droite "+"
	SndMgr.ui_createVolumeBtn("Decrease", btnMain, "RIGHT", "-")
	SndMgr.ui_createVolumeBtn("Increase", btnMain, "LEFT", "+")
	
	--------------------------------------------------------- frame texte dessous (device name)
	local labelFrame = CreateFrame("Frame", "$parent_labelFrame", SoundManagerFrame)
	labelFrame:SetPoint("TOP", SoundManagerFrame, "BOTTOM", 0, 0)
	SndMgr.setColoredFrame(labelFrame, "111111", nil, 0, addonVar.alphaUI)
	labelFrame:Hide()
	-------------------- texte
	labelFrame.text = labelFrame:CreateFontString("$parent_Text", "OVERLAY", "GameFontNormalSmall")
	SndMgr.setFontStyle(labelFrame.text, SndMgr.getStuffPath("Myriad Condensed Web.ttf", true), 11, "yellow", false, 1)
	labelFrame.text:SetPoint("CENTER", "$parent", "CENTER", 0, 0)
	labelFrame.text:SetJustifyH("CENTER")
	labelFrame.text:SetText("")--/////
end

function SndMgr.ui_initializePopupMenu() -- contenu POPUP MENU
	local frame = CreateFrame("Frame", "SoundManager_popupMenu", nil, "UIDropDownMenuTemplate")
	frame.displayMode = "MENU"
	frame.initialize = function(self, level)
		local menuContent = {}
		menuContent.padding = 10 -- /////
		
		menuContent.notCheckable = true
		menuContent.disabled = nil
		menuContent.isTitle = true
		menuContent.text = SndMgr.getLocal("Uncheck the devices to ignore")
		UIDropDownMenu_AddButton(menuContent)
		
		menuContent.notCheckable = nil
		menuContent.isTitle = nil
		menuContent.disabled = nil
		
		local currDevice = SndMgr.getCurrentDeviceName()
		for i, value in ipairs(soundDevicesTbl.active) do
			menuContent.checked = true
			menuContent.text = value.name
			if (value.name == currDevice) then 
				menuContent.colorCode = "|cff6892F2"
			else
				menuContent.colorCode = nil
			end
			menuContent.func = function(arg1, arg2) -- arg1 renvoit le nom du frame invisible et quelques propriétés
				SndMgr.ignoreDevice(arg1.value)
			end
			UIDropDownMenu_AddButton(menuContent)
		end
	
		menuContent.colorCode = nil
		for i, value in ipairs(soundDevicesTbl.ignored) do
			menuContent.checked = false
			menuContent.text = value.name
			menuContent.func = function(arg1, arg2) -- arg1 renvoit le nom du frame invisible et quelques propriétés
				SndMgr.activeDevice(arg1.value)
			end
			UIDropDownMenu_AddButton(menuContent)
		end
		
		menuContent.notCheckable = true
		menuContent.isTitle = nil
		menuContent.disabled = nil
		
		menuContent.text = SndMgr.getLocal("Set current device to default")
		menuContent.func = function() SndMgr.setDefaultDevice() end
		UIDropDownMenu_AddButton(menuContent)
		
		if (ZXSNDMGR_SAVED["settings"]["initialDevice"] == "") then menuContent.disabled = true end
		menuContent.text = SndMgr.getLocal("Reset default device")
		menuContent.func = function() SndMgr.setNoDefaultDevice() end
		UIDropDownMenu_AddButton(menuContent)
		
		menuContent.disabled = nil
		
--[[ §§§
		menuContent.text = SndMgr.getLocal("Clamp to minimap")
		menuContent.func = function() SndMgr.placeToMap() end
		UIDropDownMenu_AddButton(menuContent)
--]]
		
		menuContent.text = SndMgr.getLocal("Hide SoundManager")
		menuContent.func = function()
			SndMgr.showHide(false)
			--SndMgr.echoChatAppl(SndMgr.getLocal("hidden. Type /soundmanager to show interface again."), false, "yellow")
		end
		UIDropDownMenu_AddButton(menuContent)

		menuContent.text = SndMgr.getLocal("Display available slash commands")
		menuContent.func = function()
			SndMgr.help()
		end
		UIDropDownMenu_AddButton(menuContent)
		
		menuContent.text = SndMgr.getLocal("Cancel")
		menuContent.colorCode = "|cffAAAADD"
		menuContent.func = function() end
		UIDropDownMenu_AddButton(menuContent)
		
		menuContent.text = SndMgr.getLocal("Reset")
		menuContent.colorCode = "|cffDC571F"
		menuContent.func = function() SndMgr.resetall() end
		UIDropDownMenu_AddButton(menuContent)
	end
	return frame
end

function SndMgr.ui_createVolumeBtn(nameSuffix, parent, align, text)
	local btn = CreateFrame("Button", "SoundManager_"..nameSuffix, parent)
	btn:SetSize(addonVar.smlBtnWidth - 3, addonVar.btnHeight - 6)
	local point, relativeTo = "RIGHT", "LEFT"
	if (align == "LEFT") then
		point = "LEFT"
		relativeTo = "RIGHT"
	end
	btn:SetPoint(point, parent, relativeTo, 0, 0)
	--------------------------------- texture
	SndMgr.setColoredButton(btn, "3A52AD", nil, 0, addonVar.alphaUI, 0.5)
	--------------------------------- texte
	btn.text = btn:CreateFontString("$parent_Text", "OVERLAY", "GameFontNormalSmall")
	SndMgr.setFontStyle(btn.text, SndMgr.getStuffPath("Myriad Condensed Web.ttf", true), 13, "yellow", false, 1)
	btn.text:SetPoint("CENTER", "$parent", "CENTER", 0, 0)
	btn.text:SetJustifyH("CENTER")
	btn.text:SetText(text)--/////
	--------------------------------- script
	btn:SetScript("OnMouseUp", function(self, mouse)
		if (text == "-") then SndMgr.decreaseVolume() else SndMgr.increaseVolume() end
	end)
end

function SndMgr.ui_getBkColor(hiliteBoolean) -- renvoit la couleur RGB normal ou hilited du fond
	local r, g, b, alpha = 0.2, 0.2, 0.2, addonVar.alphaUI
	if (SndMgr.isTrue(hiliteBoolean)) then r, g, b, alpha = 0.5, 0.4, 0.4, 0.4 end
	return r, g, b, alpha
end

function SndMgr.execOutputRotation()
	local newOutputName = SndMgr.outputRotation()
	if (newOutputName ~= nil) then
		SndMgr.showDeviceNameFrame(newOutputName, false)
		
		local volumeDevice = ZXSNDMGR_SAVED["outputs"]["volumes"][newOutputName]
		local changeVolumeStr = ""
		if (volumeDevice ~= nil) then
			volumeDevice = tonumber(volumeDevice)
			if (volumeDevice ~= nil) then
				SetCVar("Sound_MasterVolume", volumeDevice)
				changeVolumeStr = ": sound volume set to "..SndMgr.convertToPercent(volumeDevice).."%"
			end
			SndMgr.updateSoundFrame(false)
		else
			SndMgr.volumeDeviceSave()
		end
		
		SndMgr.echoChatAppl("switched to "..newOutputName..changeVolumeStr, false, "green")
	else
		SndMgr.echoChatAppl("Unexpected error: unable to found device name.", false, "red")
	end
end

function SndMgr.showDeviceNameFrame(outputName, hideAfterDelayBoolean)
	local hideAfterDelayBoolean = hideAfterDelayBoolean or false
	local outputName = outputName or SndMgr.getCurrentDeviceName()
	
	SoundManagerFrame_labelFrame.text:SetText(outputName)
	local temp = SoundManagerFrame_labelFrame.text:GetStringWidth()
	SoundManagerFrame_labelFrame:SetSize(temp + 10, 12)
	SoundManagerFrame_labelFrame:Show()
	
	if (SndMgr.isTrue(hideAfterDelayBoolean)) then
		C_Timer.After(SndMgr.getVar("textdelay"), function()
			SndMgr.hideDeviceNameFrame()
		end)
	end
end

function SndMgr.hideDeviceNameFrame()
	SoundManagerFrame_labelFrame:Hide()
end

function SndMgr.updateSoundFrame(visualEffect) -- modif texte seulement si valeur différent de CVar
	if (visualEffect == nil) then visualEffect = true end
	local oldVolume = SndMgr.strToNumber(SoundManager_Switch.text:GetText())
	local currentVolume = SndMgr.convertToPercent(GetCVar("Sound_MasterVolume"))
	SndMgr.testPrint("updateSoundFrame() : visualEffect="..SndMgr.booleanToStr(visualEffect).."---oldVolume="..currentVolume.."---currentVolume="..currentVolume.."---")
	if (oldVolume ~= currentVolume) then
		SoundManager_Switch.text:SetText(currentVolume)
		SndMgr.volumeDeviceSave(currentVolume)
		if (SndMgr.isTrue(visualEffect)) then SndMgr.flash(SoundManager_Switch.texture, 0.6, 0) end -- secondes
	end
end

function SndMgr.volumeDeviceSave()
	--------------------- ménage AU DEBUT si table trop grande
	while (#(ZXSNDMGR_SAVED["outputs"]["volumes"]) > 10) do
		table.remove(ZXSNDMGR_SAVED["outputs"]["volumes"], 1)
	end
	--------------------- 
	local currentOutputDevice = SndMgr.getCurrentDeviceName()
	ZXSNDMGR_SAVED["outputs"]["volumes"][currentOutputDevice] = GetCVar("Sound_MasterVolume")
end
